#include <stdio.h>
#include <cuda.h>

#include "node.h"
#include "point.h"
#include "color.h"
#include "global.h"
#include "settings.h"
#include "real.h"

#define NODE_CACHE_SIZE 1
#define TEXTURE

__device__ __constant__ real_t STEP_SIZE_D;
__device__ __constant__ real_t INTERPOLATION_RADIUS_D;
__device__ __constant__ Ranges top_ranges;
__device__ __constant__ Node root_node;

#ifdef TEXTURE
texture<float4, cudaTextureType1D, cudaReadModeElementType> pointTexture;
#endif

inline __device__ Coord make_Coord(real_t x, real_t y, real_t z){
  Coord c = {x,y,z};
  return c;
}

inline __device__ Coord operator+(Coord a, Coord b){
  return make_Coord(a.x + b.x, a.y + b.y, a.z + b.z);
}

inline __device__ Coord operator*(Coord a, real_t s){
  return make_Coord(a.x * s, a.y * s, a.z * s);
}

__device__ Ray normalize_ray(Ray r){
  Ray t = r;
  real_t d = sqrt(r.dir.x*r.dir.x + r.dir.y*r.dir.y + r.dir.z*r.dir.z);
  t.dir.x = t.dir.x / d;
  t.dir.y = t.dir.y / d;
  t.dir.z = t.dir.z / d;

  return t;
}

__device__ Color blend_d(Color c, real_t intensity, Color* colors){
  Color new_color, output;

  new_color = colors[(int)((intensity/9.76f)*10000)];

  output.r = c.r + (1-c.a)*new_color.a*new_color.r;
  output.g = c.g + (1-c.a)*new_color.a*new_color.g;
  output.b = c.b + (1-c.a)*new_color.a*new_color.b;

  output.a = c.a + (1-c.a)*new_color.a;

  return output;
}

__device__ Display_color to_display_color_d(Color c){
  Display_color dc;

  if(c.r > 1.0f){
    c.r = 1.0;
  }

  if(c.b > 1.0f){
    c.b = 1.0;
  }

  if(c.g > 1.0f){
    c.g = 1.0;
  }
  
  dc.a = (unsigned char)(c.a*255);
  dc.r = (unsigned char)(c.r*255);
  dc.g = (unsigned char)(c.g*255);
  dc.b = (unsigned char)(c.b*255);

  return dc;
}

__device__ int get_index_d(Coord pos, Ranges r){
  char a = pos.x < r.xmin + (r.xmax - r.xmin)/2.0f;
  char b = pos.y < r.ymin + (r.ymax - r.ymin)/2.0f;
  char c = pos.z < r.zmin + (r.zmax - r.zmin)/2.0f;

  return a*4 + b*2 + c*1;
}

__device__ Ranges get_ranges_for_index_d(Ranges r, int index){
  Ranges nr = r;
  if(index == 0){
    nr.xmin += (r.xmax -r.xmin)/2.0f;
    nr.ymin += (r.ymax -r.ymin)/2.0f;
    nr.zmin += (r.zmax -r.zmin)/2.0f;
  }
  if(index == 1){
    nr.xmin += (r.xmax -r.xmin)/2.0f;
    nr.ymin += (r.ymax -r.ymin)/2.0f;
    nr.zmax -= (r.zmax -r.zmin)/2.0f;
  }
  if(index == 2){
    nr.xmin += (r.xmax -r.xmin)/2.0f;
    nr.ymax -= (r.ymax -r.ymin)/2.0f;
    nr.zmin += (r.zmax -r.zmin)/2.0f;
  }
  if(index == 3){
    nr.xmin += (r.xmax -r.xmin)/2.0f;
    nr.ymax -= (r.ymax -r.ymin)/2.0f;
    nr.zmax -= (r.zmax -r.zmin)/2.0f;
  }
  if(index == 4){
    nr.xmax -= (r.xmax -r.xmin)/2.0f;
    nr.ymin += (r.ymax -r.ymin)/2.0f;
    nr.zmin += (r.zmax -r.zmin)/2.0f;
  }
  if(index == 5){
    nr.xmax -= (r.xmax -r.xmin)/2.0f;
    nr.ymin += (r.ymax -r.ymin)/2.0f;
    nr.zmax -= (r.zmax -r.zmin)/2.0f;
  }
  if(index == 6){
    nr.xmax -= (r.xmax -r.xmin)/2.0f;
    nr.ymax -= (r.ymax -r.ymin)/2.0f;
    nr.zmin += (r.zmax -r.zmin)/2.0f;
  }
  if(index == 7){
    nr.xmax -= (r.xmax -r.xmin)/2.0f;
    nr.ymax -= (r.ymax -r.ymin)/2.0f;
    nr.zmax -= (r.zmax -r.zmin)/2.0f;
  }

  return nr;
}

__device__ real_t interpolate(Node n, Coord pos, Point* points){
  real_t weight = 0; 
  real_t intensity = 0;

  for(int c = 0; c < n.num_children; c++){
    real_t dx = pos.x - points[(long int)n.pointer + c].x;
    real_t dy = pos.y - points[(long int)n.pointer + c].y;
    real_t dz = pos.z - points[(long int)n.pointer + c].z;

    real_t distance = sqrt(dx*dx + dy*dy + dz*dz);

    if(distance < INTERPOLATION_RADIUS_D){
      intensity += (1/distance)*points[(long int)n.pointer + c].intensity;
      weight += (1/distance);
    }
  }

  if(intensity <= 0){
    return 0;
  }

  real_t ratio = intensity/weight;
  if(ratio <= 1){
    return 0;
  }

  return log(ratio);
}

inline __device__ unsigned int get_covered_subnodes(Coord pos, Node n){

  char xmin = pos.x - INTERPOLATION_RADIUS_D < n.x;
  char ymin = pos.y - INTERPOLATION_RADIUS_D < n.y;
  char zmin = pos.z - INTERPOLATION_RADIUS_D < n.z;

  char xmax = pos.x + INTERPOLATION_RADIUS_D > n.x;
  char ymax = pos.y + INTERPOLATION_RADIUS_D > n.y;
  char zmax = pos.z + INTERPOLATION_RADIUS_D > n.z;

  unsigned char b = 0;

  b = b | (unsigned int)(128 * (xmin && ymin && zmin));
  b = b | (unsigned int)(64 * (xmin && ymin && zmax));
  b = b | (unsigned int)(32 * (xmin && ymax && zmin));
  b = b | (unsigned int)(16 * (xmin && ymax && zmax));
  b = b | (unsigned int)(8 * (xmax && ymin && zmin));
  b = b | (unsigned int)(4 * (xmax && ymin && zmax));
  b = b | (unsigned int)(2 * (xmax && ymax && zmin));
  b = b | (unsigned int)(1 * (xmax && ymax && zmax));

  return b;
}

#ifdef TEXTURE
__device__ real_t get_intensity_for_pos_full(Coord pos, Node* nodes, Node* node_cache, unsigned int* stack, int base){
#else
__device__ real_t get_intensity_for_pos_full(Coord pos, Point* points, Node* nodes, Node* node_cache, unsigned int* stack, int base){
#endif

  real_t intensity = 0;
  real_t weight = 0;
  short tos = 0;
  unsigned char inc = 1;

  unsigned int current_node = 0;
  unsigned int b = get_covered_subnodes(pos, root_node);
  stack[base + tos] = current_node | (b << 24);
  tos += inc;

  while(tos > 0){
    current_node = stack[base + tos -inc] & 0x00ffffff;
    b = (stack[base + tos -inc] & 0xff000000) >> 24;

    if(b > 255){
      return 0;
    }

    unsigned int t = 1;
    while((b & t) == 0){
      t = t << 1;
    }
    current_node += (unsigned int)log2((float)t);
    b = b ^ t;

    if(b == 0){
      tos -= inc;
    }

    else{
      stack[base + tos-inc] = (stack[base + tos -inc] & 0x00ffffff) | (b << 24);
    }

    Node n;
    
    if(current_node < NODE_CACHE_SIZE){
      n = node_cache[current_node];
    }
    else{
      n = nodes[current_node];
    }

    if(n.is_leaf){
      for(short c = 0; c < n.num_children; c++){
#ifdef TEXTURE
        float4 point = tex1Dfetch(pointTexture, (long int)n.pointer + c);
#else
        Point point = points[(long int)n.pointer + c];
#endif

        real_t dx = pos.x - point.x;
        real_t dy = pos.y - point.y;
        real_t dz = pos.z - point.z;

        real_t distance = sqrt(dx*dx + dy*dy + dz*dz);

        if(distance < INTERPOLATION_RADIUS_D){
#ifdef TEXTURE
          intensity += (1/distance)*point.w;
#else
          intensity += (1/distance)*point.intensity;
#endif
          weight += (1/distance);
        }
      }
    }
    else{
      b = get_covered_subnodes(pos, nodes[current_node]);
      unsigned int temp = (unsigned int)nodes[current_node].pointer;
      stack[base + tos] = (temp) | (b << 24);
      tos += inc;
    }
  }

  if(intensity <= 0){
    return 0;
  }

  real_t ratio = intensity/weight;
  if(ratio <= 1){
    return 0;
  }

  return log(ratio);
}

__device__ real_t get_intensity_for_pos_d(Coord pos, Node* nodes, Point* points){

  Node n = root_node;
  Ranges ranges = top_ranges;
  real_t d = 0;

  while(n.is_leaf == 0){
    d++;
    int index = get_index_d(pos, ranges);
    ranges = get_ranges_for_index_d(ranges, index);

    n = nodes[((long int)n.pointer) + index];
  }

  return interpolate(n, pos, points);
}

#ifdef TEXTURE
__global__ void kernel(Node* nodes, Display_color* image, Ray* rays, Color* colors, unsigned int* stack, int stack_size){
#else
__global__ void kernel(Point* points, Node* nodes, Display_color* image, Ray* rays, Color* colors, unsigned int* stack, int stack_size){
#endif
  int i = blockIdx.x * blockDim.x + threadIdx.x;

  __shared__ Node node_cache[NODE_CACHE_SIZE];

  if(threadIdx.x < NODE_CACHE_SIZE){
    node_cache[threadIdx.x] = nodes[threadIdx.x];
  }
  __syncthreads();

  if(rays[i].distance <= 0){
    Display_color b = {0,0,0,0};
    image[i] = b;
    return;
  }

  rays[i] = normalize_ray(rays[i]);

  Coord pos = rays[i].start;
  real_t acc_distance = 0;
  Color output = {0.0,0,0,0.0};

  while(acc_distance < rays[i].distance){
#ifdef TEXTURE
    real_t intensity = get_intensity_for_pos_full(pos, nodes, node_cache, stack, i*stack_size);
#else
    real_t intensity = get_intensity_for_pos_full(pos, points, nodes, node_cache, stack, i*stack_size);
#endif

    output = blend_d(output, intensity, colors);

    if(output.a > 0.99f){
      break;
    }

    pos = pos + (rays[i].dir*STEP_SIZE_D);
    acc_distance += STEP_SIZE_D; 
  }

  image[i] = to_display_color_d(output);
}


extern "C" void launch_ray_trace_kernel(void* v){
  Tree* tree = (Tree*)v;

  cudaSetDevice(1);

  Node* nodes_d;
  Point* points_d;
  Display_color* image_d;
  Ray* rays_d;
  Color* colors_d;
  unsigned int * stack_d;
  cudaMalloc((void**)&nodes_d, sizeof(Node)*tree->node_list->current_pos);
  cudaMalloc((void**)&points_d, sizeof(Point)*tree->total_num_points);
  cudaMalloc((void**)&image_d, sizeof(Display_color)*RESOLUTION*RESOLUTION);
  cudaMalloc((void**)&rays_d, sizeof(Ray)*RESOLUTION*RESOLUTION);
  cudaMalloc((void**)&colors_d, sizeof(Color)*transfer_overlay->color_table_size);
  cudaMalloc((void**)&stack_d, sizeof(unsigned int)*RESOLUTION*RESOLUTION*10);//tree->max_depth + 2);
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));

  cudaMemcpy(nodes_d, tree->node_list->nodes, sizeof(Node)*tree->node_list->current_pos, cudaMemcpyHostToDevice);
  cudaMemcpy(points_d, tree->point_list->points, sizeof(Point)*tree->total_num_points, cudaMemcpyHostToDevice);
  cudaMemcpy(rays_d, rays, sizeof(Ray)*RESOLUTION*RESOLUTION, cudaMemcpyHostToDevice);
  cudaMemcpy(colors_d, transfer_overlay->color_table, sizeof(Color)*transfer_overlay->color_table_size, cudaMemcpyHostToDevice);
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));

#ifdef TEXTURE
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32,32,32,32, cudaChannelFormatKindFloat);
  pointTexture.filterMode = cudaFilterModePoint;
  pointTexture.normalized = false;
  cudaBindTexture(NULL, pointTexture, points_d, channelDesc, sizeof(Point)*tree->total_num_points);
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));
#endif

  cudaMemcpyToSymbol("STEP_SIZE_D", &STEP_SIZE, sizeof(real_t));
  cudaMemcpyToSymbol("INTERPOLATION_RADIUS_D", &INTERPOLATION_RADIUS, sizeof(real_t));
  cudaMemcpyToSymbol("top_ranges", &tree->ranges, sizeof(Ranges));
  cudaMemcpyToSymbol("root_node", &tree->root, sizeof(Node));
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));

  int nThreads=128;
  int totalThreads = RESOLUTION*RESOLUTION;
  int nBlocks = totalThreads/nThreads;
  nBlocks += ((totalThreads%nThreads)>0)?1:0;

#ifdef TEXTURE
  kernel<<<nBlocks, nThreads>>>(nodes_d, image_d, rays_d, colors_d, stack_d, 10);//tree->max_depth + 2);
#else
  kernel<<<nBlocks, nThreads>>>(points_d, nodes_d, image_d, rays_d, colors_d, stack_d, 10);//tree->max_depth + 2);
#endif
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));

  int images_index = log2((real_t)RESOLUTION) - 4;
  cudaMemcpy(images[images_index], image_d, sizeof(Display_color)*RESOLUTION*RESOLUTION, cudaMemcpyDeviceToHost);

  cudaFree(nodes_d);
  cudaFree(points_d);
  cudaFree(image_d);
  cudaFree(rays_d);
  cudaFree(colors_d);
  cudaFree(stack_d);
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));
  printf("%s\n", cudaGetErrorString(cudaGetLastError()));
}

extern "C" long int get_memory_size(){
  cudaDeviceProp devProp;
  cudaGetDeviceProperties(&devProp, 0);
  return devProp.totalGlobalMem;
}
